// This utility is meant to be used to ping iods to check if they're still
// alive and their UDP stack works

#define PROGNAME "iod_ping"

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <netdb.h>
extern int h_errno;

#include "iodng.h"

#define DIE(format,args...) do {		\
	fprintf(stderr, "Error: ");		\
	fprintf(stderr, format "\n",## args);	\
	exit(1);				\
} while (0)

#include "version.c"

// Gets addresses (IP and MAC) for the specified interface 
// 
// Return value: 0 if success -1 in case of failure 
// IN : itf		interface name (eth0, lo, ...)
// OUT: IPaddress	the IPv4 address (4 bytes)
// OUT: MACaddress	the HW address (6 bytes) 
//
static int get_if_addr(unsigned char *IPaddress,
	       	unsigned char *MACaddress,
	        const char *const itf) {
	struct ifreq ifr;
	int fd;

	fd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (fd == -1) {
		printf("cannot open a socket errno=%d\n", errno);
		goto error;
	}

	strncpy (ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFADDR, &ifr)) {
		printf ("cannot get the IP address of %s (SIOCGIFADDR)", ifr.ifr_name);
		goto error;
	}
	memcpy(IPaddress, &(((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr), 4);

	strncpy (ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFHWADDR, &ifr)) {
		printf ("cannot get the HW address of %s (SIOCGIFHWADDR)",
			ifr.ifr_name);
		goto error;
	}
	memcpy(MACaddress, ifr.ifr_hwaddr.sa_data, 6);

	close (fd);

	return 0;
	
error:
	memset(IPaddress, '\0', 4);
	memset(MACaddress, '\0', 6);
	return -1;
}


// Resolve the hostname 'host' and fills sin
// 
// OUT: psin		the socket structure to fill
// IN : host		the hostname
// IN : port		the port
//
static void inet_resolve(struct sockaddr_in *psin,
		const char * const host, const char * const port) {

	struct hostent *he;
	int numport;

	memset(psin, '\0', sizeof(*psin));


	psin->sin_family = AF_INET;

	he = gethostbyname(host);
	if (he != NULL) {
		memcpy((char*)&(psin->sin_addr), he->h_addr, he->h_length);
	} else {
		DIE("cannot resolve host %s (errno=%d)", host, errno);
	}

	numport = atoi(port);
	if (0 < numport && numport < 65536) {
		psin->sin_port = htons(numport);
	} else {
		DIE("port must be between 1 and 65535 and is currently %d", numport);
	}
	psin->sin_port = htons(numport);
}


// Displays usage
static void usage(FILE* f) {
	fprintf(f,	
"Usage: " PROGNAME " [OPTIONS] host[:port]\n"
"\n"
"Try and pings the iod on host:port\n"
"\n"
" -h, --help               print this terse help screen\n"
" -v, --version            print the version and exit\n"
" -s h[:p], --spoof h[:p]  iod will spoof host:port h:p     (def: no spoof)\n"
" -c n, --count n          send exactly n pings            (def: unlimited)\n"
" -w n.nn, --wait n.nn     wait n.nn second(s) between each ping   (def: 2)\n"
" -d, --debug              enable verbosity              (unused currently)"
"\n"
	);
	exit((f==stderr)?1:0);
}

static float wait = 2.0;		// number of seconds between each packet
static int count = -1;			// number of packets
static int spoofing = 0;		// boolean value
static struct sockaddr_in spoof;	// valid if (spoofing)
static struct sockaddr_in iod;		//
static int verbosity = 0;

// Parse the command line
static void parse_command_line(int argc, char *argv[]) {
	static const char * shortopts = "hvs:c:w:d";
	static struct option longopts[] = {
		{ "help",	0,			0, 'h' },
		{ "version",	0,			0, 'v' },
		{ "spoof",	required_argument,	0, 's' },
		{ "count",	required_argument,	0, 'c' },
		{ "wait",	required_argument,	0, 'w' },
		{ "debug",	0,			0, 'd' },
		{ NULL,		0,			0, 0 }
	};
	char c;

	opterr = 0;
	while((c = getopt_long(argc, argv, shortopts, longopts, NULL)) != EOF) {
		switch(c) {
			case 'h':
				usage(stdout);
				break;
			case 'v':
				printf("iodng - %s\n", version);
				exit(0);
				break;
			case 's':
				do {
					char *spoofh, *spoofp;

					spoofing = 1;
					spoofh = optarg;
					spoofp = strchr(spoofh, ':');
					if (NULL == spoofp) {
						spoofp = "33333";
					} else {
						*spoofp++ = '\0';
					}
					inet_resolve(&spoof, spoofh, spoofp);
				} while (0);
				break;
			case 'c':
				count = atoi(optarg);
				if (count < 1) {
					DIE("count cannot be smaller than 1: %s", optarg);
				}
				break;
			case 'w':
				wait = atof(optarg);
				if (wait < 0) {
					DIE("wait may not be negative: %s", optarg);
				}
				break;
			case 'd':
				verbosity++;
				break;
			default:
				fprintf(stderr, "Error: unknown option\n");
				usage(stderr);
		}
	}

	if (argc - optind == 1) {
		char *iodh, *iodp;
		iodh = argv[optind];
		iodp = strchr(iodh, ':');
		if (NULL == iodp) {
			iodp = "20000";
		} else {
			*iodp++ = '\0';
		}
		inet_resolve(&iod, iodh, iodp);
	} else {
		DIE("incorrect number of arguments\n");
	}
}

// Initialize the random numbers generator used to get a random seed
void init_rand() {
	struct timeval tv;

	if (-1 == gettimeofday(&tv, NULL)) {
		DIE("could not get current time (gettimeofday() failed errno=%d)", errno);
	}
	srand(tv.tv_usec);
}

// Main
int main(int argc, char *argv[]) {
	int sockfd;
	int failed_reply = 0;
	msg_iod_t ping_req, pong_rep;

	struct sockaddr_in reply_sin;
	socklen_t reply_sin_len;
	unsigned char eth0_ip[4];
	unsigned char eth0_mac[6];

	parse_command_line(argc, argv);

	init_rand();


	// Get a socket descriptor
	sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sockfd < 0) {
		DIE("could not open a UDP socket (socket() failed errno=%d)", errno);
	}

	if (-1 == get_if_addr(eth0_ip, eth0_mac, "eth0")) {
		DIE("could not run get_if_addr()");
	} else {
		printf("Using interface eth0 (IP: %u.%u.%u.%u MAC:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n",
				eth0_ip[0], eth0_ip[1], eth0_ip[2], eth0_ip[3],
				eth0_mac[0], eth0_mac[1],eth0_mac[2],eth0_mac[3],eth0_mac[4],eth0_mac[5]);
	}


	printf("IOD_PING %s:%d with %d bytes of data\n",
			inet_ntoa(iod.sin_addr), ntohs(iod.sin_port),
			MSG_IOD_PING_LEN);

	ping_req.pingreq.seq = rand() & 0xFFFF0000;
	while (0 != count) {
		int sent, rcvd, ret;
		fd_set rfds;
		struct timeval tv;
		struct timeval tv_ping, tv_pong;
		int latency;

		if (count > 0) {
			count--;
		}

		// Sending PING REQUEST

		ping_req.magic = NFSP_MAGIC_MSG;
		ping_req.msgtype = NFSP_PING_REQ;
		ping_req.pingreq.seq++;

		ping_req.client_addr = 0;	// 0 => the iod will fill it
		ping_req.client_port = 0;	// 0 => the iod will fill it

		memcpy(ping_req.client_mac_addr, eth0_mac, 6);

		ping_req.server_addr = (spoofing) ? spoof.sin_addr.s_addr : 0;
		ping_req.server_port = (spoofing) ? spoof.sin_port : 0;

		// Send the pingreq through sockfd
		sent = sendto(sockfd, (char*) &ping_req, MSG_IOD_PING_LEN, MSG_NOSIGNAL, 
					(struct sockaddr *)&iod, sizeof(struct sockaddr_in));
		if (sent == -1) {
			fprintf(stderr, "Error: could not sendto (errno=%d)\n", errno);

			if (count) {
				usleep((int) (wait * 1000000.0));
			}
			continue;
		}

		// Waiting PONG_REPLY

		FD_ZERO(&rfds);
		FD_SET(sockfd, &rfds);
		tv.tv_sec = 0;
		tv.tv_usec = (int) (wait * 1000000.0);

		ret = gettimeofday(&tv_ping, NULL);
		if (ret) {
			DIE("gettimeofday failed() (errno = %d)", errno);
		}

		ret = select(sockfd + 1, &rfds, NULL, NULL, &tv);
		switch(ret) {
			case -1:
				DIE("could not select() (errno=%d)",errno);
				break;

			case 0:
				// Timeout
				printf("Reply for seq=0x%08lx: FAILED (timeout)\n",
				   	ping_req.pingreq.seq);
				failed_reply = 1;
				break;

			default:
				reply_sin_len = sizeof(reply_sin);
				rcvd = recvfrom(sockfd,(char*)&pong_rep, MSG_IOD_PING_LEN, 0,
					   	(struct sockaddr *)&reply_sin, &reply_sin_len);

				ret = gettimeofday(&tv_pong, NULL);
				if (ret) {
					DIE("gettimeofday failed() (errno = %d)", errno);
				}
				latency = 1000000 * (tv_pong.tv_sec - tv_ping.tv_sec) + tv_pong.tv_usec - tv_ping.tv_usec;

				if (ping_req.pingreq.seq == pong_rep.pingreq.seq 
					&& pong_rep.magic == NFSP_MAGIC_MSG 
					&& pong_rep.msgtype == NFSP_PONG_REQ 
					&& rcvd == MSG_IOD_PING_LEN) {
					printf("Reply from %s:%d : seq=0x%08lx time=%d.%d ms\n",
					   	inet_ntoa(reply_sin.sin_addr), ntohs(reply_sin.sin_port),
					   	ping_req.pingreq.seq, latency / 1000, latency % 1000);
				} else {
					printf("Reply from %s:%d for seq=0x%08lx: FAILED"
							" (seq=0x%08lx magic=0x%04x msgtype=0x%04x "
							"sent=%d rcvd=%d) time=%d.%d ms\n",
						   	inet_ntoa(reply_sin.sin_addr), ntohs(reply_sin.sin_port),
							ping_req.pingreq.seq, pong_rep.pingreq.seq,
							pong_rep.magic, pong_rep.msgtype,
						       	sent, rcvd, latency / 1000, latency % 1000);
				}
				if (count) {
					int to_sleep;		//in microseconds
					struct timespec req, rem;

					to_sleep = (int) ((1000000.0 * wait)) - latency;
					if (to_sleep < 0) {
						fprintf(stderr, "Warning: requesting negative time to sleep (%dus) => "
								" forced to 0us\n", to_sleep);
						to_sleep = 0;
					}
					
					memset(&req, 0, sizeof(req));
					req.tv_sec = to_sleep / 1000000;
					req.tv_nsec = (to_sleep % 1000000) * 1000;
					memset(&rem, 0, sizeof(rem));
					ret = nanosleep(&req, &rem);
					if (ret) {
						DIE("nanosleep failed() (errno = %d)", errno);
					}
				}
		}
	}

	return (failed_reply) ? -1 : 0;
}
